# ai_hydra/constants/DPong.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from typing import Final


class DPLabel:
    TITLE: Final[str] = "HydraPong"


class DPField:
    BOX: Final[str] = "box"
    PONG_BOARD: Final[str] = "pong_board"
    PONG_BOX: Final[str] = "pong_box"
    TITLE: Final[str] = "title"


class DPFile:
    PONG_CSS: Final[str] = "HydraPong.tcss"
